from __future__ import annotations

from typing import TYPE_CHECKING

from django_twc_toolbox.conf import app_settings

if TYPE_CHECKING:
    from sentry_sdk._types import SamplingContext


def sentry_traces_sampler(sampling_context: SamplingContext):
    if _should_disgard(sampling_context):
        return 0

    return app_settings.SENTRY_TRACES_RATE


def sentry_profiles_sampler(sampling_context: SamplingContext):
    if _should_disgard(sampling_context):
        return 0

    return app_settings.SENTRY_PROFILE_RATE


def _should_disgard(sampling_context: SamplingContext) -> bool:
    DISGARDED_METHODS = app_settings.SENTRY_DISGARDED_METHODS
    DISGARDED_PATHS = app_settings.SENTRY_DISGARDED_PATHS

    return (
        sampling_context.get("wsgi_environ", None) is not None
        and sampling_context["wsgi_environ"]["REQUEST_METHOD"] in DISGARDED_METHODS
        and sampling_context["wsgi_environ"]["PATH_INFO"] in DISGARDED_PATHS
    )



from __future__ import annotations

from django.http import HttpRequest
from django.http import HttpResponse
from django.shortcuts import render
from django.utils import timezone
from django.views.decorators.cache import cache_control
from django.views.decorators.http import require_GET

from .conf import app_settings


def custom_error_404(
    request: HttpRequest,
    exception: Exception | None = None,
    *args: object,
    **kwargs: object,
) -> HttpResponse:
    return render(request, app_settings.TEMPLATE_404, context={}, status=404)


def custom_error_500(
    request: HttpRequest, *args: object, **kwargs: object
) -> HttpResponse:
    return render(request, app_settings.TEMPLATE_500, context={}, status=500)


@require_GET
@cache_control(max_age=app_settings.CACHE_TIME_ROBOTS_TXT, immutable=True, public=True)
def robots_txt(request: HttpRequest) -> HttpResponse:
    return render(request, app_settings.TEMPLATE_ROBOTS_TXT, content_type="text/plain")


@require_GET
@cache_control(
    max_age=app_settings.CACHE_TIME_SECURITY_TXT, immutable=True, public=True
)
def security_txt(request: HttpRequest) -> HttpResponse:
    return render(
        request,
        app_settings.TEMPLATE_SECURITY_TXT,
        context={
            "year": timezone.now().year + 1,
        },
        content_type="text/plain",
    )



# pyright: reportDeprecated=false
# ruff: noqa: UP006
import enum
from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Mapping
from typing import ClassVar
from typing import List
from typing import TypeVar
from typing import override

from django import forms
from django.core.files.uploadedfile import UploadedFile
from django.core.paginator import Page
from django.core.paginator import Paginator
from django.db import models
from django.http import HttpRequest
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.template.response import TemplateResponse
from django.urls import URLPattern
from django.utils.datastructures import MultiValueDict
from django.utils.decorators import classonlymethod
from django.utils.functional import classproperty
from django.views.generic import View

class Role(enum.Enum):
    LIST = "list"
    DETAIL = "detail"
    CREATE = "create"
    UPDATE = "update"
    DELETE = "delete"
    def handlers(self) -> dict[str, str]: ...
    def extra_initkwargs(self) -> dict[str, str]: ...
    @property
    def url_name_component(self) -> str: ...
    def url_pattern(self, view_cls: type[View]) -> str: ...
    def get_url(self, view_cls: type[View]) -> URLPattern: ...
    def reverse(self, view: CRUDView, object: models.Model | None = None) -> str: ...
    def maybe_reverse(
        self, view: CRUDView, object: models.Model | None = None
    ) -> str | None: ...

_TModel = TypeVar("_TModel", bound=models.Model)
_TObject = object

class CRUDView(View):
    role: Role
    model: ClassVar[type[models.Model] | None] = None
    fields: ClassVar[List[str] | None] = None

    lookup_field: ClassVar[str]
    lookup_url_kwarg: ClassVar[str | None] = None
    path_converter: ClassVar[str]
    object: models.Model | None = None
    object_list: Iterable[models.Model] | None = None

    queryset: ClassVar[models.QuerySet[models.Model] | None] = None
    form_class: ClassVar[forms.Form | None] = None
    template_name: ClassVar[str | None] = None
    context_object_name: ClassVar[str | None] = None

    paginate_by: ClassVar[int | None] = None
    page_kwarg: ClassVar[str] = "page"
    allow_empty: ClassVar[bool] = True

    template_name_suffix: ClassVar[str | None] = None

    def list(
        self, request: HttpRequest, *args: _TObject, **kwargs: _TObject
    ) -> TemplateResponse: ...
    def detail(
        self, request: HttpRequest, *args: _TObject, **kwargs: _TObject
    ) -> TemplateResponse: ...
    def show_form(
        self, request: HttpRequest, *args: _TObject, **kwargs: _TObject
    ) -> TemplateResponse: ...
    def process_form(
        self, request: HttpRequest, *args: _TObject, **kwargs: _TObject
    ) -> HttpResponse | HttpResponseRedirect: ...
    def confirm_delete(
        self, request: HttpRequest, *args: _TObject, **kwargs: _TObject
    ) -> TemplateResponse: ...
    def process_deletion(
        self, request: HttpRequest, *args: _TObject, **kwargs: _TObject
    ) -> TemplateResponse: ...
    def get_queryset(self) -> models.QuerySet[models.Model]: ...
    def get_object(self) -> models.Model: ...
    def get_form_class(self) -> type[forms.Form]: ...
    def get_form(
        self,
        data: Mapping[str, _TObject] | None = None,
        files: MultiValueDict[str, UploadedFile] | None = None,
        **kwargs: _TObject,
    ) -> forms.Form: ...
    def form_valid(self, form: forms.Form) -> HttpResponseRedirect: ...
    def form_invalid(self, form: forms.Form) -> HttpResponse: ...
    def get_success_url(self) -> str: ...
    def get_paginate_by(self) -> int | None: ...
    def get_paginator(
        self, queryset: models.QuerySet[_TModel], page_size: int
    ) -> Paginator[_TModel]: ...
    def paginate_queryset(
        self, queryset: models.QuerySet[_TModel], page_size: int
    ) -> Page[_TModel]: ...
    def get_filterset(
        self,
        queryset: models.QuerySet[models.Model] | None = None,
    ) -> _TObject: ...  # TODO: change Any to FilterSet
    def get_context_object_name(self, is_list: bool = False) -> str | None: ...
    def get_context_data(self, **kwargs: _TObject) -> dict[str, _TObject]: ...
    def get_template_names(self) -> List[str]: ...
    def render_to_response(
        self, context: dict[str, _TObject] | None = None
    ) -> TemplateResponse: ...
    @override
    @classmethod
    def as_view(  # type: ignore[override]
        cls, role: Role, **initkwargs: _TObject
    ) -> Callable[..., HttpResponse]: ...
    @classproperty
    def url_base(cls) -> str: ...
    @classonlymethod
    def get_urls(cls, roles: List[Role] | None = None) -> List[URLPattern]: ...



from __future__ import annotations

import logging
from pathlib import Path

import pytest
from django.conf import settings
from django.http import HttpRequest

from .settings import DEFAULT_SETTINGS

pytest_plugins = []  # type: ignore


def pytest_configure(config):
    logging.disable(logging.CRITICAL)

    settings.configure(**DEFAULT_SETTINGS, **TEST_SETTINGS)


TEST_SETTINGS = {
    "INSTALLED_APPS": [
        "django_twc_toolbox",
        "django_twc_toolbox.crud",
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django_tables2",
        "simple_history",
        "template_partials",
        "tests.dummy",
        "tests.test_crud",
    ],
    "TEMPLATES": [
        {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "DIRS": [Path(__file__).parent / "templates"],
            "OPTIONS": {
                "context_processors": [
                    "django.template.context_processors.request",
                ],
                "loaders": [
                    (
                        "template_partials.loader.Loader",
                        [
                            "django.template.loaders.filesystem.Loader",
                            "django.template.loaders.app_directories.Loader",
                        ],
                    )
                ],
            },
        }
    ],
}


@pytest.fixture
def req():
    return HttpRequest()



from typing import ClassVar

from django_tables2 import tables

TableData = dict[str, object]

class TableMixinBase:
    context_table_name: ClassVar[str]
    table_pagination: bool | dict[str, object] | None
    def get_paginate_by(self, table_data: TableData | None) -> int | None: ...

class SingleTableMixin(TableMixinBase):
    table_class: ClassVar[type[tables.Table] | None] = None
    table_data: ClassVar[TableData | None] = None
    def get_context_data(self, **kwargs: object) -> dict[str, object]: ...



from __future__ import annotations

from django.db import models

from django_twc_toolbox.models import WithHistory


class DateOrderableModel(models.Model):
    date = models.DateField()


class DateTimeOrderableModel(models.Model):
    date = models.DateTimeField()


class ModelWithHistory(WithHistory):
    name = models.CharField(max_length=255)


class Parent(models.Model):
    foo = models.CharField(max_length=255)


class Child(models.Model):
    parent = models.ForeignKey(
        "dummy.Parent", on_delete=models.CASCADE, related_name="parents"
    )
    bar = models.CharField(max_length=255)



from __future__ import annotations

import pytest
from django.test import override_settings

from django_twc_toolbox.conf import app_settings
from django_twc_toolbox.sentry import _should_disgard
from django_twc_toolbox.sentry import sentry_profiles_sampler
from django_twc_toolbox.sentry import sentry_traces_sampler


@pytest.fixture
def sample_context():
    return {"wsgi_environ": {"REQUEST_METHOD": "GET", "PATH_INFO": "/health/"}}


@pytest.fixture
def non_discard_context():
    return {"wsgi_environ": {"REQUEST_METHOD": "POST", "PATH_INFO": "/api/data/"}}


def test_sentry_traces_sampler_discard(sample_context):
    assert sentry_traces_sampler(sample_context) == 0


def test_sentry_traces_sampler_non_discard(non_discard_context):
    assert sentry_traces_sampler(non_discard_context) == app_settings.SENTRY_TRACES_RATE


def test_sentry_profiles_sampler_discard(sample_context):
    assert sentry_profiles_sampler(sample_context) == 0


def test_sentry_profiles_sampler_non_discard(non_discard_context):
    assert (
        sentry_profiles_sampler(non_discard_context) == app_settings.SENTRY_PROFILE_RATE
    )


def test_should_discard(sample_context):
    assert _should_disgard(sample_context) is True


def test_should_not_discard(non_discard_context):
    assert _should_disgard(non_discard_context) is False


def test_should_discard_empty_context():
    assert _should_disgard({}) is False


@override_settings(
    DJANGO_TWC_TOOLBOX={
        "SENTRY_DISGARDED_METHODS": ["POST"],
        "SENTRY_DISGARDED_PATHS": ["/api/"],
        "SENTRY_TRACES_RATE": 0.75,
        "SENTRY_PROFILE_RATE": 0.25,
    }
)
def test_custom_settings():
    context = {"wsgi_environ": {"REQUEST_METHOD": "POST", "PATH_INFO": "/api/"}}

    assert _should_disgard(context) is True
    assert (
        sentry_traces_sampler(
            {"wsgi_environ": {"REQUEST_METHOD": "GET", "PATH_INFO": "/other/"}}
        )
        == 0.75
    )
    assert (
        sentry_profiles_sampler(
            {"wsgi_environ": {"REQUEST_METHOD": "GET", "PATH_INFO": "/other/"}}
        )
        == 0.25
    )


@pytest.mark.parametrize(
    "method,path,expected",
    [
        ("GET", "/health/", True),
        ("HEAD", "/health/", True),
        ("POST", "/health/", False),
        ("GET", "/api/", False),
        ("HEAD", "/api/", False),
    ],
)
def test_should_discard_parametrized(method, path, expected):
    context = {"wsgi_environ": {"REQUEST_METHOD": method, "PATH_INFO": path}}

    assert _should_disgard(context) == expected



from __future__ import annotations

from collections.abc import Iterable
from typing import TypeVar

from django import template
from django.contrib.auth.models import AbstractUser
from django.contrib.auth.models import AnonymousUser
from django.core.paginator import Page
from django.core.paginator import Paginator
from django.db import models
from django.utils.itercompat import is_iterable

register = template.Library()


@register.filter
def initials(user: AbstractUser) -> str:
    if not hasattr(user, "username") or isinstance(user, AnonymousUser):
        return "N/A"

    initials = user.username[0].upper()

    if user.first_name and user.last_name:
        initials = f"{user.first_name[0].upper()}{user.last_name[0].upper()}"

    return initials


@register.filter
def display_name(user: AbstractUser) -> str:
    if not hasattr(user, "username") or isinstance(user, AnonymousUser):
        return "N/A"

    if user.first_name and user.last_name:
        display_name = f"{user.first_name} {user.last_name}"
    elif user.first_name:
        display_name = user.first_name
    else:
        display_name = user.username

    return display_name.title()


_TModel = TypeVar("_TModel", bound=models.Model)


@register.simple_tag()
def elided_page_range(page_obj: Page[_TModel]) -> Iterable[str | int]:
    paginator: Paginator[_TModel] | None = getattr(page_obj, "paginator", None)
    if not paginator:
        return []
    return paginator.get_elided_page_range(page_obj.number)


# Coming in the next Django version
# Copied from https://github.com/django/django/pull/17368
# TODO: Remove this when Django 5.1 is released
@register.simple_tag(takes_context=True)
def query_string(context, query_dict=None, **kwargs):
    """
    Add, remove, and change parameters of a ``QueryDict`` and return the result
    as a query string. If the ``query_dict`` argument is not provided, default
    to ``request.GET``.
    For example::
        {% query_string foo=3 %}
    To remove a key::
        {% query_string foo=None %}
    To use with pagination::
        {% query_string page=page_obj.next_page_number %}
    A custom ``QueryDict`` can also be used::
        {% query_string my_query_dict foo=3 %}
    """
    if query_dict is None:
        query_dict = context.request.GET
    query_dict = query_dict.copy()
    for key, value in kwargs.items():
        if value is None:
            if key in query_dict:
                del query_dict[key]
        elif is_iterable(value) and not isinstance(value, str):
            query_dict.setlist(key, value)
        else:
            query_dict[key] = value
    if not query_dict:
        return ""
    query_string = query_dict.urlencode()
    return f"?{query_string}"


_T = TypeVar("_T")


@register.filter
def klass(instance: _T) -> type[_T]:
    return instance.__class__


@register.filter
def class_name(instance: object) -> str:
    return instance.__class__.__name__



from __future__ import annotations

import sys
from typing import TypeVar

from django.db.models.base import Model

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:  # pragma: no cover
    from typing_extensions import TypeAlias  # pyright: ignore[reportUnreachable]


_K = TypeVar("_K")

if sys.version_info >= (3, 10):
    ListOrTuple: TypeAlias = list[_K] | tuple[_K, ...] | tuple[()]
else:
    from typing import List
    from typing import Tuple
    from typing import Union

    ListOrTuple: TypeAlias = Union[List[_K], Tuple[_K, ...], Tuple[()]]

ChildModelT = TypeVar("ChildModelT", bound=Model)
ParentModelT = TypeVar("ParentModelT", bound=Model)


if sys.version_info >= (3, 12):
    from typing import override as typing_override
else:  # pragma: no cover
    from typing_extensions import (
        override as typing_override,  # pyright: ignore[reportUnreachable]
    )

override = typing_override



from __future__ import annotations

import pytest
from model_bakery import baker
from neapolitan.views import Role

from .models import Bookmark
from .views import BookmarkTableView
from .views import BookmarkView


@pytest.mark.parametrize(
    "klass,expected",
    [
        (BookmarkView, False),
        (BookmarkTableView, True),
    ],
)
def test_rendered_template_table(klass, expected, rf, db):
    view_func = klass.as_view(role=Role.LIST)
    request = rf.get(Role.LIST.maybe_reverse(klass))
    baker.make(Bookmark, _quantity=3)

    response = view_func(request)

    assert response.status_code == 200

    rendered = response.render()
    content = rendered.content.decode()

    assert ("<h1>With Table</h1>" in content) is expected


def test_rendered_partial_template(rf, db):
    view_func = BookmarkView.as_view(role=Role.LIST)
    request = rf.get(Role.LIST.maybe_reverse(BookmarkView))
    request.htmx = True
    baker.make(Bookmark, _quantity=3)

    response = view_func(request)
    rendered = response.render()
    content = rendered.content.decode()

    # this div is just outside where the partial is defined in the default list template
    assert '<div class="sm:flex sm:items-center">' not in content


@pytest.mark.xfail(
    reason="have not figured out how to dynamically change the partial name using `partialdef` templatetag"
)
def test_rendered_partial_template_different_target(rf, db):
    view_func = BookmarkView.as_view(
        role=Role.LIST, list_partial="test_different_target"
    )
    request = rf.get(Role.LIST.maybe_reverse(BookmarkView))
    request.htmx = True
    baker.make(Bookmark, _quantity=3)

    response = view_func(request)
    rendered = response.render()
    content = rendered.content.decode()

    # this div is just outside where the partial is defined in the default list template
    assert '<div class="sm:flex sm:items-center">' not in content



from __future__ import annotations

import datetime
import itertools
import random
from dataclasses import asdict
from dataclasses import dataclass

import pytest
from django import VERSION as DJANGO_VERSION
from django.core.paginator import EmptyPage
from django.core.paginator import Page
from django.core.paginator import PageNotAnInteger
from django.db.models.query import QuerySet
from django.utils import timezone
from model_bakery import baker

from django_twc_toolbox.paginator import DatePage
from django_twc_toolbox.paginator import DatePaginator

from .dummy.models import DateOrderableModel
from .dummy.models import DateTimeOrderableModel


@dataclass
class ModelClassParams:
    """
    Stores parameters for a pytest parameterized test case.

    Intended to be used as a parameter to a parameterized test case along
    with the indirect `model_data_queryset` fixture.
    """

    model_class: type[DateOrderableModel | DateTimeOrderableModel]
    number_of_days: int

    def __iter__(self):
        return iter(asdict(self).values())

    @classmethod
    def from_request(cls, request):
        return cls(
            model_class=request.param.model_class,
            number_of_days=request.param.number_of_days,
        )


@pytest.fixture
def model_data_queryset(request, db):
    """
    Sets up a queryset of the specified model class with the specified number of days of data
    for use in pagination tests.

    Depending on the test parameterization, it uses `ModelClassParams` to determine the model class
    and the number of days for data generation. If not parameterized with `ModelClassParams`,
    it defaults to `DateOrderableModel` with 90 days of data.

    The fixture supports creating data for both `DateOrderableModel` and `DateTimeOrderableModel`.
    For `DateTimeOrderableModel`, it generates two data points per day.

    The DatePaginator expects an ordered data structure, so the data is returned ordered to make
    all the tests do not need to worry about the initial ordering. In the tests that need to verify
    the ordering, they can apply their own sorting to the data.
    """
    # Check if the request has the expected parameterization
    if hasattr(request, "param") and isinstance(request.param, ModelClassParams):
        model_class, number_of_days = ModelClassParams.from_request(request)
    else:
        # Default values for tests that don't use ModelClassParams
        model_class = DateOrderableModel  # or any default model
        number_of_days = 90  # default number of days

    if model_class == DateOrderableModel:
        date_values = (
            timezone.now() - datetime.timedelta(days=i) for i in range(number_of_days)
        )
    else:
        current_time = timezone.now()
        date_values = itertools.chain.from_iterable(
            (
                current_time - datetime.timedelta(days=i, hours=12),
                current_time - datetime.timedelta(days=i),
            )
            for i in range(number_of_days)
        )

    baker.make(
        model_class,
        date=itertools.cycle(date_values),
        _quantity=number_of_days,
    )

    return model_class.objects.all().order_by("date")


@pytest.fixture(params=["queryset", "list", "tuple"])
def objects(request, model_data_queryset):
    """
    Provides the model data needed in different formats to make sure our custom
    DatePaginator can handle all the data structures a Django Paginator should handle.

    It uses the `model_data_queryset` fixture to get the initial queryset and then
    provides it in three different structures: QuerySet, list, and tuple.
    """
    if request.param == "queryset":
        ret = model_data_queryset
    elif request.param == "tuple":
        ret = tuple(model_data_queryset)
    else:
        ret = list(model_data_queryset)

    return ret


class TestDatePaginator:
    @pytest.mark.parametrize(
        "model_data_queryset,days_per_page,expected_num_pages",
        [
            # date
            # create 90 days of data, paginate by 30 days, expect 3 pages
            (
                ModelClassParams(model_class=DateOrderableModel, number_of_days=90),
                30,
                3,
            ),
            # datetime
            # create 90 days of data (2 per day), paginate by 30 days, expect 2 pages
            (
                ModelClassParams(model_class=DateTimeOrderableModel, number_of_days=90),
                30,
                2,
            ),
            # date
            # create 180 days of data, paginate by 30 days, expect 6 pages
            (
                ModelClassParams(model_class=DateOrderableModel, number_of_days=180),
                30,
                6,
            ),
            # datetime
            # create 180 days of data (2 per day), paginate by 30 days, expect 3 pages
            (
                ModelClassParams(
                    model_class=DateTimeOrderableModel, number_of_days=180
                ),
                30,
                3,
            ),
            # date
            # create 90 days of data, paginate by 60 days, expect 2 pages
            (
                ModelClassParams(model_class=DateOrderableModel, number_of_days=90),
                60,
                2,
            ),
            # date
            # create 90 days of data (2 per day), paginate by 60 days, expect 1 page
            (
                ModelClassParams(model_class=DateTimeOrderableModel, number_of_days=90),
                60,
                1,
            ),
        ],
        indirect=["model_data_queryset"],
    )
    def test_num_pages(self, objects, days_per_page, expected_num_pages):
        paginator = DatePaginator(
            objects, "date", datetime.timedelta(days=days_per_page)
        )

        assert paginator.num_pages == expected_num_pages

    def test_first_page(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=30))
        first_page = paginator.page(1)

        first_segment_start, first_segment_end = paginator.date_segments[0]

        # Count objects in the first segment based on the type of 'objects'
        if isinstance(objects, QuerySet):  # type: ignore[misc]
            objects_in_first_segment = objects.filter(
                date__gte=first_segment_start,
                date__lt=first_segment_end,
            ).count()
        else:
            objects_in_first_segment = sum(
                1
                for entry in objects
                if first_segment_start <= entry.date < first_segment_end
            )

        assert len(first_page.object_list) == objects_in_first_segment

    def test_page_out_of_range(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=30))

        with pytest.raises(EmptyPage):
            paginator.page(4)  # Only 3 pages exist

    def test_with_empty_queryset(self):
        empty = DateOrderableModel.objects.none().order_by("date")
        paginator = DatePaginator(empty, "date", datetime.timedelta(days=30))

        assert paginator.count == 0
        assert paginator.num_pages == 0

    def test_with_empty_list(self):
        paginator = DatePaginator([], "date", datetime.timedelta(days=30))

        assert paginator.count == 0
        assert paginator.num_pages == 0

    def test_with_single_entry(self, db):
        baker.make("dummy.DateOrderableModel", date=timezone.now().date())
        obj = DateOrderableModel.objects.all().order_by("-date")
        paginator = DatePaginator(obj, "date", datetime.timedelta(days=30))

        assert paginator.num_pages == 1
        assert len(paginator.page(1).object_list) == 1

    def test_large_date_range(self, objects):
        large_range = datetime.timedelta(days=365)
        paginator = DatePaginator(objects, "date", large_range)

        assert paginator.num_pages == 1
        assert len(paginator.page(1).object_list) == len(objects)

    def test_small_date_range(self, objects):
        small_range = datetime.timedelta(days=1)
        paginator = DatePaginator(objects, "date", small_range)

        assert paginator.num_pages == len(objects)

    def test_date_range_boundary_cases(self, db, objects):
        if isinstance(objects, QuerySet):  # type: ignore[misc]
            earliest_date = objects.earliest("date").date
        else:
            earliest_date = min(entry.date for entry in objects)

        paginator = DatePaginator(objects, "date", datetime.timedelta(days=30))
        start_of_first_segment, _ = paginator.date_segments[0]

        # Ensure the boundary date falls within the first segment
        boundary_date = max(earliest_date, start_of_first_segment)

        # Create new objects at the boundary date
        new_objects = baker.make(
            "dummy.DateOrderableModel", date=boundary_date, _quantity=5
        )

        # If 'objects' is a tuple/list, manually add these new objects to it
        if not isinstance(objects, QuerySet):  # type: ignore[misc]
            if isinstance(objects, tuple):
                objects = objects + tuple(new_objects)
            else:
                objects.extend(new_objects)

        first_page = paginator.page(1)

        assert all(obj.date >= boundary_date for obj in first_page.object_list)

    def test_non_sequence_page_number(self, objects):
        if isinstance(objects, QuerySet):  # type: ignore[misc]
            random_objects = objects.order_by("?")
        else:
            random_objects = list(objects)
            random.shuffle(random_objects)
            if isinstance(objects, tuple):
                random_objects = tuple(random_objects)

        with pytest.raises(ValueError):
            DatePaginator(random_objects, "date", datetime.timedelta(days=30))

    @pytest.mark.parametrize(
        "model_data_queryset",
        [
            ModelClassParams(model_class=DateOrderableModel, number_of_days=90),
            ModelClassParams(model_class=DateTimeOrderableModel, number_of_days=180),
        ],
        indirect=["model_data_queryset"],
    )
    def test_chronological_ordering(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=30))

        first_page_item_date = paginator.page(1).object_list[0].date
        middle_page_item_date = paginator.page(2).object_list[0].date
        last_page_item_date = paginator.page(3).object_list[0].date

        assert first_page_item_date < middle_page_item_date
        assert middle_page_item_date < last_page_item_date

        for item in paginator.page(1).object_list:
            assert item.date >= first_page_item_date

    @pytest.mark.parametrize(
        "model_data_queryset",
        [
            ModelClassParams(model_class=DateOrderableModel, number_of_days=90),
            ModelClassParams(model_class=DateTimeOrderableModel, number_of_days=180),
        ],
        indirect=["model_data_queryset"],
    )
    def test_reversed_ordering(self, objects):
        if isinstance(objects, QuerySet):  # type: ignore[misc]
            objects = objects.order_by("-date")
        else:
            objects = list(reversed(objects))

        paginator = DatePaginator(objects, "date", datetime.timedelta(days=30))

        first_page_item_date = paginator.page(1).object_list[0].date
        middle_page_item_date = paginator.page(2).object_list[0].date
        last_page_item_date = paginator.page(3).object_list[0].date

        assert first_page_item_date > middle_page_item_date
        assert middle_page_item_date > last_page_item_date

        previous_item = first_page_item_date
        for item in paginator.page(1).object_list:
            assert item.date <= first_page_item_date
            assert item.date <= previous_item
            previous_item = item.date

    def test_paginator_page_function(self, objects):
        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))

        page = paginator.page(1)

        assert isinstance(page, Page)
        assert page.number == 1

        with pytest.raises(PageNotAnInteger):
            paginator.page("not_a_number")

        with pytest.raises(EmptyPage):
            paginator.page(0)  # Page number less than 1

        with pytest.raises(EmptyPage):
            paginator.page(paginator.num_pages + 1)  # Page number out of range

    @pytest.mark.parametrize(
        "model_data_queryset",
        [
            ModelClassParams(model_class=DateOrderableModel, number_of_days=90),
            ModelClassParams(model_class=DateTimeOrderableModel, number_of_days=180),
        ],
        indirect=["model_data_queryset"],
    )
    def test_paginator_date_range(self, objects):
        date_range = datetime.timedelta(days=10)
        paginator = DatePaginator(objects, "date", date_range)

        with pytest.warns(DeprecationWarning):
            assert paginator.date_range == date_range

    @pytest.mark.parametrize(
        "model_data_queryset",
        [
            ModelClassParams(model_class=DateOrderableModel, number_of_days=90),
            ModelClassParams(model_class=DateTimeOrderableModel, number_of_days=180),
        ],
        indirect=["model_data_queryset"],
    )
    def test_paginator_page_date_range(self, objects):
        date_range = datetime.timedelta(days=10)
        paginator = DatePaginator(objects, "date", date_range)

        assert paginator.page_date_range == date_range

    @pytest.mark.parametrize(
        "model_data_queryset",
        [
            ModelClassParams(model_class=DateOrderableModel, number_of_days=90),
            ModelClassParams(model_class=DateTimeOrderableModel, number_of_days=180),
        ],
        indirect=["model_data_queryset"],
    )
    def test_paginator_page_date_range_missing(self, objects):
        with pytest.raises(ValueError):
            DatePaginator(objects, "date")

    @pytest.mark.parametrize(
        "model_data_queryset",
        [
            ModelClassParams(model_class=DateOrderableModel, number_of_days=90),
            ModelClassParams(model_class=DateTimeOrderableModel, number_of_days=180),
        ],
        indirect=["model_data_queryset"],
    )
    def test_paginator_explicit_date_range(self, objects):
        date_range = datetime.timedelta(days=10)

        with pytest.warns(DeprecationWarning):
            DatePaginator(objects, "date", date_range=date_range)


class TestDatePaginatorInheritance:
    """
    Test that the DatePaginator respects the Django Paginator inheritance.

    Only test the attributes and methods that are not overridden.
    """

    @pytest.fixture
    def paginator(self, objects):
        return DatePaginator(objects, "date", datetime.timedelta(days=10))

    def test___init__(self, objects):
        paginator = DatePaginator(
            objects,
            "date",
            datetime.timedelta(days=10),
            allow_empty_first_page=False,
        )

        assert paginator.object_list == objects
        assert paginator.per_page == 1  # As it paginates by date
        assert paginator.allow_empty_first_page is False
        if DJANGO_VERSION >= (5, 0, 0):
            assert paginator.error_messages == paginator.default_error_messages

    def test_orphans(self, objects):
        with pytest.warns(UserWarning):
            DatePaginator(objects, "date", datetime.timedelta(days=10), orphans=5)

    def test_allow_empty_first_page(self):
        allow_paginator = DatePaginator(
            DateOrderableModel.objects.none().order_by("date"),
            "date",
            datetime.timedelta(days=10),
            allow_empty_first_page=True,
        )

        assert allow_paginator.num_pages == 0

        do_not_allow_paginator = DatePaginator(
            DateOrderableModel.objects.none().order_by("date"),
            "date",
            datetime.timedelta(days=10),
            allow_empty_first_page=False,
        )

        assert do_not_allow_paginator.num_pages == 0
        with pytest.raises(EmptyPage):
            do_not_allow_paginator.page(1)

    @pytest.mark.skipif(
        DJANGO_VERSION < (5, 0, 0),
        reason="`error_messages` kwarg introduced in Django 5.0",
    )
    def test_error_messages(self, objects):
        custom_error_messages = {
            "invalid_page": "Custom invalid page message",
            "min_page": "Custom min page message",
            "no_results": "Custom no results message",
        }

        paginator = DatePaginator(
            objects,
            "date",
            datetime.timedelta(days=10),
            error_messages=custom_error_messages,
        )

        with pytest.raises(PageNotAnInteger) as excinfo:
            paginator.page("invalid")
        assert str(excinfo.value) == custom_error_messages["invalid_page"]

        with pytest.raises(EmptyPage) as excinfo:
            paginator.page(0)
        assert str(excinfo.value) == custom_error_messages["min_page"]

        with pytest.raises(EmptyPage) as excinfo:
            paginator.page(1000)
        assert str(excinfo.value) == custom_error_messages["no_results"]

    def test___iter__(self, paginator):
        for index, page in enumerate(paginator):
            assert page.number == index + 1

    def test_validate_number(self, paginator):
        assert paginator.validate_number(1) == 1

        with pytest.raises(PageNotAnInteger):
            paginator.validate_number("two")

        with pytest.raises(EmptyPage):
            paginator.validate_number(-1)

        with pytest.raises(EmptyPage):
            paginator.validate_number(paginator.num_pages + 1)

    def test_get_page(self, paginator):
        page = paginator.get_page(1)

        assert isinstance(page, Page)
        assert page.number == 1

        # Invalid page number
        invalid_page = paginator.get_page(1000)

        assert invalid_page.number == paginator.num_pages  # Should return the last page

    def test_count(self, objects):
        paginator = DatePaginator(
            objects,
            "date",
            datetime.timedelta(days=10),
        )

        assert paginator.count == len(objects)

    def test_page_range(self, paginator):
        assert len(paginator.page_range) == paginator.num_pages

    def test_get_elided_page_range(self, paginator):
        elided_range = paginator.get_elided_page_range(number=1)
        elided_range_list = list(elided_range)

        assert all(
            isinstance(item, int) or item == paginator.ELLIPSIS
            for item in elided_range_list
        )


class TestDatePage:
    @pytest.mark.parametrize(
        "order_by",
        ["date", "-date"],
    )
    def test_page_start_date(self, objects, order_by):
        if isinstance(objects, QuerySet):  # type: ignore[misc]
            objects = objects.order_by(order_by)
        elif order_by == "date":
            objects = list(objects)
        else:
            objects = list(reversed(objects))

        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        start_date, _ = paginator.date_segments[0]

        assert page.start_date == start_date

    @pytest.mark.parametrize(
        "order_by",
        ["date", "-date"],
    )
    def test_page_start_date_single_entry(self, db, order_by):
        baker.make("dummy.DateOrderableModel", date=timezone.now().date())
        obj = DateOrderableModel.objects.all().order_by(order_by)

        paginator = DatePaginator(obj, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        start_date, _ = paginator.date_segments[0]

        assert page.start_date == start_date

    @pytest.mark.parametrize(
        "order_by",
        ["date", "-date"],
    )
    def test_page_end_date(self, objects, order_by):
        if isinstance(objects, QuerySet):  # type: ignore[misc]
            objects = objects.order_by(order_by)
        elif order_by == "date":
            objects = list(objects)
        else:
            objects = list(reversed(objects))

        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        _, end_date = paginator.date_segments[0]

        assert page.end_date == end_date

    @pytest.mark.parametrize(
        "order_by",
        ["date", "-date"],
    )
    def test_page_end_date_single_entry(self, db, order_by):
        baker.make("dummy.DateOrderableModel", date=timezone.now().date())
        obj = DateOrderableModel.objects.all().order_by(order_by)

        paginator = DatePaginator(obj, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        _, end_date = paginator.date_segments[0]

        assert page.end_date == end_date

    @pytest.mark.parametrize(
        "order_by",
        ["date", "-date"],
    )
    def test_page_min_date(self, objects, order_by):
        if isinstance(objects, QuerySet):  # type: ignore[misc]
            objects = objects.order_by(order_by)
        elif order_by == "date":
            objects = list(objects)
        else:
            objects = list(reversed(objects))

        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        start_date, end_date = paginator.date_segments[0]

        if order_by == "date":
            assert page.min_date == start_date
        else:
            assert page.min_date == end_date

    @pytest.mark.parametrize(
        "order_by",
        ["date", "-date"],
    )
    def test_page_max_date(self, objects, order_by):
        if isinstance(objects, QuerySet):  # type: ignore[misc]
            objects = objects.order_by(order_by)
        elif order_by == "date":
            objects = list(objects)
        else:
            objects = list(reversed(objects))

        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        start_date, end_date = paginator.date_segments[0]

        if order_by == "date":
            assert page.max_date == end_date
        else:
            assert page.max_date == start_date

        page = paginator.page(2)

        start_date, end_date = paginator.date_segments[1]

        if order_by == "date":
            assert page.max_date == end_date
        else:
            assert page.max_date == start_date

    @pytest.mark.parametrize(
        "order_by",
        ["date", "-date"],
    )
    def test_page_date_range(self, objects, order_by):
        if isinstance(objects, QuerySet):  # type: ignore[misc]
            objects = objects.order_by(order_by)
        elif order_by == "date":
            objects = list(objects)
        else:
            objects = list(reversed(objects))

        paginator = DatePaginator(objects, "date", datetime.timedelta(days=10))
        page = paginator.page(1)

        start_date, end_date = paginator.date_segments[0]

        if order_by == "date":
            assert page.date_range == (start_date, end_date)
        else:
            assert page.date_range == (end_date, start_date)


class TestDatePageInheritence:
    """
    Test that the DatePage respects the Django Page inheritance.

    Only test the attributes and methods that are not overridden.
    """

    @pytest.fixture
    def paginator(self, objects):
        return DatePaginator(objects, "date", datetime.timedelta(days=10))

    def test___init__(self, paginator):
        page = paginator.page(1)
        start_date, end_date = paginator.date_segments[0]

        date_page = DatePage(
            page.object_list, page.number, paginator, start_date, end_date
        )

        assert date_page.object_list == page.object_list
        assert date_page.number == page.number
        assert date_page.paginator == page.paginator

    def test___len__(self, paginator):
        page = paginator.page(1)
        start_date, end_date = paginator.date_segments[0]

        date_page = DatePage(
            page.object_list, page.number, paginator, start_date, end_date
        )

        assert len(date_page) == len(page.object_list)

    def test___getitem__(self, paginator):
        page = paginator.page(1)
        start_date, end_date = paginator.date_segments[0]

        date_page = DatePage(
            page.object_list, page.number, paginator, start_date, end_date
        )

        assert date_page[0] == page.object_list[0]
        assert date_page[:2] == page.object_list[:2]
        with pytest.raises(IndexError):
            date_page[len(page.object_list)]

    def test_has_next(self, paginator):
        last_page_number = paginator.num_pages
        last_start_date, last_end_date = paginator.date_segments[last_page_number - 1]

        last_page = DatePage(
            paginator.page(last_page_number).object_list,
            last_page_number,
            paginator,
            last_start_date,
            last_end_date,
        )

        assert not last_page.has_next()

        second_last_page_number = last_page_number - 1
        second_last_start_date, second_last_end_date = paginator.date_segments[
            second_last_page_number - 1
        ]

        second_last_page = DatePage(
            paginator.page(second_last_page_number).object_list,
            second_last_page_number,
            paginator,
            second_last_start_date,
            second_last_end_date,
        )

        assert second_last_page.has_next()

    def test_has_previous(self, paginator):
        first_start_date, first_end_date = paginator.date_segments[0]

        first_page = DatePage(
            paginator.page(1).object_list,
            1,
            paginator,
            first_start_date,
            first_end_date,
        )

        assert not first_page.has_previous()

        second_start_date, second_end_date = paginator.date_segments[1]

        second_page = DatePage(
            paginator.page(2).object_list,
            2,
            paginator,
            second_start_date,
            second_end_date,
        )

        assert second_page.has_previous()

    def test_has_other_pages(self, objects, paginator):
        first_start_date, first_end_date = paginator.date_segments[0]

        first_page = DatePage(
            paginator.page(1).object_list,
            1,
            paginator,
            first_start_date,
            first_end_date,
        )

        assert first_page.has_other_pages()

        single_paginator = DatePaginator(
            objects, "date", datetime.timedelta(days=len(objects) + 1)
        )
        single_start_date, single_end_date = single_paginator.date_segments[0]

        single_page = DatePage(
            single_paginator.page(1).object_list,
            1,
            single_paginator,
            single_start_date,
            single_end_date,
        )

        assert not single_page.has_other_pages()

    def test_next_page_number(self, paginator):
        first_start_date, first_end_date = paginator.date_segments[0]

        first_page = DatePage(
            paginator.page(1).object_list,
            1,
            paginator,
            first_start_date,
            first_end_date,
        )

        assert first_page.next_page_number() == 2

        last_page_number = paginator.num_pages
        last_start_date, last_end_date = paginator.date_segments[last_page_number - 1]

        last_page = DatePage(
            paginator.page(last_page_number).object_list,
            last_page_number,
            paginator,
            last_start_date,
            last_end_date,
        )

        with pytest.raises(EmptyPage):
            last_page.next_page_number()

    def test_previous_page_number(self, paginator):
        last_page_number = paginator.num_pages
        last_start_date, last_end_date = paginator.date_segments[last_page_number - 1]

        last_page = DatePage(
            paginator.page(last_page_number).object_list,
            last_page_number,
            paginator,
            last_start_date,
            last_end_date,
        )

        assert last_page.previous_page_number() == last_page_number - 1

        first_start_date, first_end_date = paginator.date_segments[0]

        first_page = DatePage(
            paginator.page(1).object_list,
            1,
            paginator,
            first_start_date,
            first_end_date,
        )

        with pytest.raises(EmptyPage):
            first_page.previous_page_number()

    def test_start_index(self, paginator):
        first_start_date, first_end_date = paginator.date_segments[0]

        first_page = DatePage(
            paginator.page(1).object_list,
            1,
            paginator,
            first_start_date,
            first_end_date,
        )

        assert first_page.start_index() == 1

        middle_page_number = paginator.num_pages // 2
        middle_start_date, middle_end_date = paginator.date_segments[
            middle_page_number - 1
        ]

        middle_page = DatePage(
            paginator.page(middle_page_number).object_list,
            middle_page_number,
            paginator,
            middle_start_date,
            middle_end_date,
        )

        assert (
            middle_page.start_index()
            == (middle_page_number - 1) * paginator.per_page + 1
        )

    def test_end_index(self, paginator):
        first_start_date, first_end_date = paginator.date_segments[0]

        first_page = DatePage(
            paginator.page(1).object_list,
            1,
            paginator,
            first_start_date,
            first_end_date,
        )

        assert first_page.end_index() == min(
            first_page.number * paginator.per_page, paginator.count
        )

        last_page_number = paginator.num_pages
        last_start_date, last_end_date = paginator.date_segments[last_page_number - 1]

        last_page = DatePage(
            paginator.page(last_page_number).object_list,
            last_page_number,
            paginator,
            last_start_date,
            last_end_date,
        )

        assert last_page.end_index() == paginator.count



from __future__ import annotations

import os
from unittest.mock import patch

import pytest
from django.conf import settings
from django.core.management import CommandError
from django.core.management import call_command
from django.test.utils import override_settings
from model_bakery import baker

pytestmark = pytest.mark.django_db


@override_settings(DEBUG=True)
def test_createsuperuser():
    user = baker.make("User", is_superuser=True)

    with patch.dict(os.environ, {"DJANGO_SUPERUSER_PASSWORD": "password"}):
        call_command("createsuperuser", interactive=False, username=user.username)

    user.refresh_from_db()

    assert user.check_password("password")


@override_settings(DEBUG=True)
def test_createsuperuser_update_email():
    user = baker.make("User", is_superuser=True, email="test@example.com")

    with patch.dict(os.environ, {"DJANGO_SUPERUSER_PASSWORD": "password"}):
        call_command(
            "createsuperuser",
            interactive=False,
            username=user.username,
            email="updated@example.com",
        )

    user.refresh_from_db()

    assert user.email == "updated@example.com"


@override_settings(DEBUG=False)
def test_createsuperuser_not_debug():
    user = baker.make("User", is_superuser=True)

    with patch.dict(os.environ, {"DJANGO_SUPERUSER_PASSWORD": "password"}):
        with pytest.raises(CommandError):
            call_command("createsuperuser", interactive=False, username=user.username)


@override_settings(
    INSTALLED_APPS=[
        app for app in settings.INSTALLED_APPS if app != "django_twc_toolbox"
    ]
)
def test_createsuperuser_no_toolbox():
    user = baker.make("User", is_superuser=True)

    with patch.dict(os.environ, {"DJANGO_SUPERUSER_PASSWORD": "password"}):
        with pytest.raises(CommandError):
            call_command("createsuperuser", interactive=False, username=user.username)



# Taken from https://github.com/softwarecrafts/django-cptemplate
#
# MIT License
#
# Copyright (c) 2024 Andrew Miller
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from pathlib import Path
from shutil import copy2

from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand
from django.template import loader


def get_template_absolute_path(template_path):
    try:
        template = loader.get_template(template_path)
        return template.origin.name  # type: ignore[attr-defined]
    except Exception as e:
        print(f"Error occurred while getting template path: {e}")
        return None


class Command(BaseCommand):
    help = "Copies a template from a package into your project"

    def add_arguments(self, parser):
        parser.add_argument("source", type=str)
        parser.add_argument("destination", type=str, nargs="?")

    def handle(self, *args, **options):
        print(options)
        # 1. extract the options
        source = options["source"]
        destination = options.get("destination")
        # 2. Check that the source file exists - use template loaders for this
        source_file = get_template_absolute_path(source)
        if source_file is None:
            self.stdout.write(self.style.ERROR("Source Template doesn't exist"))
            return
        base_dir = Path(settings.BASE_DIR)  # type: ignore[misc]
        # 4. if destination, then create Path object and copy
        if destination is not None:
            app_config = apps.get_app_config(destination)
            destination_path = base_dir / app_config.path / "templates" / source
        # 5/ else inspect TEMPLATES[DIRS] setting and use first option if available
        else:
            try:
                template_dir = Path(settings.TEMPLATES[0]["DIRS"][0])
                destination_path = template_dir / source
            except IndexError:
                # 6/ otherwise create project level template directory and dump file there.
                destination_path = base_dir / "templates" / source
                # PRINT what settings need to be modified
                self.stdout.write(
                    self.style.WARNING(
                        'Update TEMPLATES["DIRS"] to include the following entry "BASE_DIR / "templates","'
                    )
                )

        destination_path.parent.mkdir(parents=True, exist_ok=True)
        copy2(source_file, destination_path)



from __future__ import annotations

from functools import partial

from charidfield import CharIDField
from cuid import cuid

CuidField = partial(
    CharIDField,
    default=cuid,
    max_length=30,
    help_text="cuid-format identifier for this entity.",
)



# pyright: reportUnnecessaryTypeIgnoreComment=false
from __future__ import annotations

import sys
from collections.abc import Callable
from typing import ClassVar
from typing import Literal

from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.http import Http404
from django.http import HttpRequest
from django.http import HttpResponse
from django.template.response import TemplateResponse
from django_htmx.middleware import HtmxDetails
from django_tables2 import tables
from django_tables2.views import SingleTableMixin
from neapolitan.views import CRUDView as NeapolitanCRUDView
from neapolitan.views import Role

if sys.version_info >= (3, 12):
    from typing import override
else:  # pragma: no cover
    from typing_extensions import override  # pyright: ignore[reportUnreachable]


class HtmxHttpRequest(HttpRequest):
    htmx: HtmxDetails | None


class CRUDView(NeapolitanCRUDView):
    paginate_by = 100

    detail_fields: ClassVar[list[str] | None] = None
    list_fields: ClassVar[list[str] | None] = None

    table_class: ClassVar[type[tables.Table] | None] = None
    table_data: ClassVar[dict[str, object] | None] = None

    # django-template-partials doesn't seem to be able to use a passed in context
    # variable to define partials, e.g. with `list_partial = "object-list"`
    # in the template context:
    #
    # ```htmldjango
    # {% load partials %}
    # {% partialdef list_partial %}
    #   <h1>Hello World!</h1>
    # {% endpartialdef %}
    # ```
    #
    # This does not work, at least the different ways I've tried.
    #
    # However! I am including this class variable in the event I (or someone else)
    # cracks the case and figures out how to get the templatetag to be able to take
    # an outside variable.
    #
    # So the partial name within the `object_list.html` template MUST BE `object-list`,
    # at least for the time being.
    list_partial: ClassVar[Literal["object-list"]] = "object-list"

    filterset_primary_fields: list[str] | None = None

    request: HtmxHttpRequest  # pyright: ignore[reportIncompatibleVariableOverride]

    def get_fields(self):
        match self.role:
            case Role.DETAIL:
                fields = self.get_detail_fields()
            case Role.LIST:
                fields = self.get_list_fields()
            case _:
                fields = None

        if fields is not None:
            return fields

        if self.fields is not None:
            return self.fields

        msg = "'%s' must define 'fields' or override 'get_fields()'"
        raise ImproperlyConfigured(msg % self.__class__.__name__)

    def get_detail_fields(self):
        return self.detail_fields

    def get_list_fields(self):
        return self.list_fields

    @override
    def list(
        self, request: HttpRequest, *args: object, **kwargs: object
    ) -> TemplateResponse:
        """GET handler for the list view."""

        queryset = self.get_queryset()

        filterset = self.get_filterset(queryset)
        if filterset is not None:
            queryset = filterset.qs  # type:ignore[attr-defined]

        if not self.allow_empty and not queryset.exists():
            raise Http404

        self.object_list = queryset

        paginate_by = self.get_paginate_by(self.object_list)

        if paginate_by is None:
            context = self.get_context_data(
                page_obj=None,
                is_paginated=False,
                paginator=None,
                filterset=filterset,
            )
        else:
            page = self.paginate_queryset(self.object_list, paginate_by)
            context = self.get_context_data(
                page_obj=page,
                is_paginated=page.has_other_pages(),
                paginator=page.paginator,
                filterset=filterset,
            )

        return self.render_to_response(context)

    @override
    def get_paginate_by(self, *args: object, **kwargs: object) -> int | None:
        return super().get_paginate_by()

    @override
    def get_filterset(
        self, queryset: models.QuerySet[models.Model] | None = None
    ) -> object | None:
        filterset = super().get_filterset(queryset)

        if filterset is None:
            return None

        if self.filterset_primary_fields is not None:
            filterset.primary_fields = self.filterset_primary_fields  # type: ignore[attr-defined]
            filterset.secondary_fields = list(  # type: ignore[attr-defined]
                set(filterset.form.fields.keys()) - set(filterset.primary_fields)  # type:ignore[attr-defined]
            )

        def is_active(filterset: object) -> bool:
            return any(
                value not in (None, "", [], {})
                for value in filterset.form.cleaned_data.values()  # type: ignore[attr-defined]
            )

        def get_active_filters(filterset: object):
            return {
                key: value
                for key, value in filterset.form.cleaned_data.items()  # type:ignore[attr-defined]
                if value not in (None, "", [], {})
            }

        filterset.__class__.is_active = is_active  # type:ignore[attr-defined]
        filterset.__class__.active_filters = property(get_active_filters)  # type:ignore[attr-defined]

        return filterset

    @override
    def get_context_data(self, **kwargs: object) -> dict[str, object]:
        context = super().get_context_data(**kwargs)
        context["list_view_url"] = Role.LIST.maybe_reverse(self)
        if self.object is not None:
            context["delete_view_url"] = Role.DELETE.maybe_reverse(self, self.object)
            context["detail_view_url"] = Role.DETAIL.maybe_reverse(self, self.object)
            context["update_view_url"] = Role.UPDATE.maybe_reverse(self, self.object)
        context.update(self.get_role_context_data(context, **kwargs))
        return context

    def get_role_context_data(
        self, context: dict[str, object], **kwargs: object
    ) -> dict[str, object]:
        if not hasattr(self, "role"):
            return {}

        func_name = f"get_{self.role.value}_context_data"
        func = getattr(self, func_name, None)

        if func is None or not callable(func):
            return {}

        role_context = func(context=context, **kwargs)

        if not isinstance(role_context, dict):
            msg = f"`{func_name}` must return a dictionary"
            raise ValueError(msg)

        return role_context

    @override
    def get_template_names(self):
        template_names = super().get_template_names()

        # only render the template partial if:
        # - it's the list view
        # - it's an HTMX request
        # - it's not a paginated request
        #
        # Right now, our custom templates do not have support for rendering template partials when
        # dealing with paginated lists. We could probably update it to add an `hx-target` to the
        # pagination links, but for now we'll just render the entire template.
        if (
            self.role == Role.LIST
            and getattr(self.request, "htmx", False)
            and self.request.htmx
            and not self.kwargs.get(self.page_kwarg, None)  # pyright: ignore[reportAny]
            and not self.request.GET.get(self.page_kwarg, None)
        ):
            template_names = [
                f"{template_name}#{self.list_partial}"
                for template_name in template_names
            ]

        return template_names

    @classmethod
    @override
    def as_view(  # type: ignore[override]
        cls, role: Role, **initkwargs: object
    ) -> Callable[..., HttpResponse]:
        # Check if the list view is being called OR the list view is being called and there is no
        # `table_class` attribute set. If not, we can just return the parent
        # `neapolitan.views.CRUDView.as_view` method to render as normal.

        if role != Role.LIST or cls.table_class is None:
            return super().as_view(role=role, **initkwargs)

        # View is a list view and has the `table_class` attribute set, so we need to override the class
        # returned by adding `django_tables2.views.SingleTableMixin` so that the table can be rendered.
        # I would normally just add this to the base CRUDView up top, but since this class is used for
        # multiple request types, we just scope it to the GET on the list url. Is this overkill? Would this
        # be a bit clearer if it was just set above in the class declaration instead of doing this dance?
        # Unsure. Open to changing this later on in case this is too complicated.
        #
        # We don't need to change anything else about the class (there's already some small logic around
        # dealing with pagination in the `list` method above), so we can get away with just adding the mixin,
        # inheriting from `cls` a.k.a. `django_twc_toolbox.crud.views.CRUDView`, and leaving the body of the
        # new view        # class empty (`...`).
        #
        # Also need to pass in the class variable `table_class` to the `as_view` class method so it's available
        # on the instance.

        class ListViewWithTable(SingleTableMixin, cls): ...  # type: ignore[misc,valid-type]

        return ListViewWithTable.as_view(
            role=role, table_class=cls.table_class, **initkwargs
        )



from __future__ import annotations

import pytest
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.core.paginator import Paginator
from django.http import QueryDict
from django.template import Context
from django.template import RequestContext
from django.template import Template
from model_bakery import baker

from django_twc_toolbox.templatetags.django_twc_toolbox import class_name
from django_twc_toolbox.templatetags.django_twc_toolbox import display_name
from django_twc_toolbox.templatetags.django_twc_toolbox import elided_page_range
from django_twc_toolbox.templatetags.django_twc_toolbox import initials
from django_twc_toolbox.templatetags.django_twc_toolbox import klass
from django_twc_toolbox.templatetags.django_twc_toolbox import query_string

pytestmark = pytest.mark.django_db

User = get_user_model()


@pytest.mark.parametrize(
    "model,kwargs,expected",
    [
        (User, {"username": "johndoe", "first_name": "John", "last_name": "Doe"}, "JD"),
        (User, {"username": "janedoe"}, "J"),
        (AnonymousUser, {}, "N/A"),
    ],
)
def test_initials(model, kwargs, expected):
    user = model(**kwargs)

    assert initials(user) == expected


@pytest.mark.parametrize(
    "model,kwargs,expected",
    [
        (User, {"username": "johndoe", "first_name": "John", "last_name": "Doe"}, "JD"),
        (User, {"username": "janedoe"}, "J"),
        (AnonymousUser, {}, "N/A"),
    ],
)
def test_initials_templatetag(model, kwargs, expected):
    if kwargs:
        user = baker.make(model, **kwargs)
    else:
        user = model()

    template = Template("{% load django_twc_toolbox %} Initials: {{ user|initials }}")

    rendered = template.render(Context({"user": user}))

    assert f"Initials: {expected}" in rendered


@pytest.mark.parametrize(
    "model,kwargs,expected",
    [
        (
            User,
            {"username": "johndoe", "first_name": "John", "last_name": "Doe"},
            "John Doe",
        ),
        (User, {"username": "janedoe", "first_name": "Jane"}, "Jane"),
        (User, {"username": "bobsmith"}, "Bobsmith"),
        (AnonymousUser, {}, "N/A"),
    ],
)
def test_display_name(model, kwargs, expected):
    user = model(**kwargs)

    assert display_name(user) == expected


@pytest.mark.parametrize(
    "model,kwargs,expected",
    [
        (
            User,
            {"username": "johndoe", "first_name": "John", "last_name": "Doe"},
            "John Doe",
        ),
        (User, {"username": "janedoe", "first_name": "Jane"}, "Jane"),
        (User, {"username": "bobsmith"}, "Bobsmith"),
        (AnonymousUser, {}, "N/A"),
    ],
)
def test_display_name_templatetag(model, kwargs, expected):
    if kwargs:
        user = baker.make(model, **kwargs)
    else:
        user = model()

    template = Template("{% load django_twc_toolbox %} Name: {{ user|display_name }}")

    rendered = template.render(Context({"user": user}))

    assert f"Name: {expected}" in rendered


@pytest.mark.parametrize(
    "object_count,per_page,page,expected",
    [
        (101, 5, 1, [1, 2, 3, 4, Paginator.ELLIPSIS, 20, 21]),
        (
            101,
            5,
            10,
            [
                1,
                2,
                Paginator.ELLIPSIS,
                7,
                8,
                9,
                10,
                11,
                12,
                13,
                Paginator.ELLIPSIS,
                20,
                21,
            ],
        ),
        (101, 5, 21, [1, 2, Paginator.ELLIPSIS, 18, 19, 20, 21]),
    ],
)
def test_elided_page_range(object_count, per_page, page, expected):
    objects = range(object_count)
    paginator = Paginator(objects, per_page=per_page)
    page_obj = paginator.page(page)

    result = list(elided_page_range(page_obj))
    assert result == expected


@pytest.mark.parametrize(
    "object_count,per_page,page,expected",
    [
        (101, 5, 1, f"Pages: 1 2 3 4 {Paginator.ELLIPSIS} 19 20"),
        (
            101,
            5,
            10,
            f"Pages: 1 2 {Paginator.ELLIPSIS} 7 8 9 10 11 12 13 {Paginator.ELLIPSIS} 19 20",
        ),
        (101, 5, 20, f"Pages: 1 2 {Paginator.ELLIPSIS} 17 18 19 20"),
    ],
)
def test_elided_page_range_templatetag(object_count, per_page, page, expected):
    objects = range(1, object_count)
    paginator = Paginator(objects, per_page=per_page)
    page_obj = paginator.page(page)

    template = Template(
        "{% load django_twc_toolbox %} {% elided_page_range page_obj as range %} Pages: {% for p in range %}{{ p }} {% endfor %}"
    )

    rendered = template.render(Context({"page_obj": page_obj}))

    assert expected in rendered


@pytest.mark.parametrize(
    "url,params,expected",
    [
        ("/test/?page=2&sort=name", {"foo": "bar"}, "?page=2&sort=name&foo=bar"),
        ("/test/?page=2&sort=name", {"page": 3}, "?page=3&sort=name"),
        ("/test/?page=2&sort=name", {"sort": None}, "?page=2"),
        ("/test/?page=2&sort=name", {"page": None, "sort": None}, ""),
    ],
)
def test_query_string(url, params, expected, rf):
    request = rf.get(url)

    result = query_string(RequestContext(request), **params)

    assert result == expected


def test_query_string_querydict(rf):
    request = rf.get("/test/?page=2&sort=name")
    custom_query = QueryDict("a=1&b=2&c=3", mutable=True)

    result = query_string(RequestContext(request), custom_query, b="new", d="added")

    assert result == "?a=1&b=new&c=3&d=added"


@pytest.mark.parametrize(
    "url,params,expected",
    [
        (
            "/test/?page=2&sort=name",
            {"foo": "bar"},
            "?page=2&amp;sort=name&amp;foo=bar",
        ),
        ("/test/?page=2&sort=name", {"page": 3}, "?page=3&amp;sort=name"),
        ("/test/?page=2&sort=name", {"sort": None}, "?page=2"),
        ("/test/?page=2&sort=name", {"page": None, "sort": None}, ""),
    ],
)
def test_query_string_templatetag(url, params, expected, rf):
    request = rf.get(url)

    param_string = " ".join(
        [
            f"{key}='{value}'" if value is not None else f"{key}=None"
            for key, value in params.items()
        ]
    )
    template = Template(
        f"{{% load django_twc_toolbox %}} Query: {{% query_string {param_string} %}}"
    )

    rendered = template.render(RequestContext(request, params))

    assert f"Query: {expected}" in rendered


def test_klass():
    template = Template("")

    result = klass(template)

    assert result == Template


def test_klass_templatetag():
    template = Template("{% load django_twc_toolbox %} {{ template|klass }}")

    rendered = template.render(Context({"template": template}))

    assert "django.template.base.Template" in rendered


def test_class_name():
    template = Template("")

    result = class_name(template)

    assert result == "Template"


def test_class_name_templatetag():
    template = Template("{% load django_twc_toolbox %} {{ template|class_name }}")

    rendered = template.render(Context({"template": template}))

    assert "Template" in rendered



# type: ignore
from __future__ import annotations

import os

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.management.commands import createsuperuser
from django.core.management import CommandError


class Command(createsuperuser.Command):
    help = "Create or update superuser."

    def handle(self, *args, **options):
        try:
            super().handle(*args, **options)
        except CommandError as err:
            if "That username is already taken" not in str(err) or not settings.DEBUG:
                raise err

            User = get_user_model()

            username = options[User.USERNAME_FIELD]
            user = User.objects.get(username=username)

            password = os.environ["DJANGO_SUPERUSER_PASSWORD"]
            user.set_password(password)

            if email := options[User.EMAIL_FIELD]:
                user.email = email

            user.save()



from __future__ import annotations

from model_bakery import baker
from neapolitan.views import Role

from django_twc_toolbox.crud.templatetags.neapolitan import action_links
from django_twc_toolbox.crud.templatetags.neapolitan import object_detail
from django_twc_toolbox.crud.templatetags.neapolitan import object_list

from .models import Bookmark
from .test_views import BookmarkView


def test_action_links(db):
    view = BookmarkView()
    object = baker.make(Bookmark)

    actions = action_links(view, object)

    assert actions["detail"]["url"] == f"/bookmark/{object.pk}/"
    assert actions["update"]["url"] == f"/bookmark/{object.pk}/edit/"
    assert actions["delete"]["url"] == f"/bookmark/{object.pk}/delete/"


def test_object_detail(db):
    view = BookmarkView(role=Role.DETAIL)
    object = baker.make(Bookmark)

    detail = object_detail(object, view)
    detail_list = list(detail["object"])

    assert detail_list[0][0] == view.detail_fields[0]
    assert detail_list[0][1] == getattr(object, view.detail_fields[0])
    assert detail_list[1][0] == view.detail_fields[1]
    assert detail_list[1][1] == getattr(object, view.detail_fields[1])


def test_object_list(db):
    view = BookmarkView(role=Role.LIST)
    objects = baker.make(Bookmark, _quantity=5)

    list = object_list(objects, view)

    assert list["headers"] == view.list_fields
    assert len(list["object_list"]) == len(objects)

    detail = list["object_list"][0]["fields"]

    assert detail[0]["value"] == getattr(objects[0], view.list_fields[0])



from __future__ import annotations

import pytest
from django.contrib import admin
from model_bakery import baker

from django_twc_toolbox.admin import ReadOnlyStackedInline
from django_twc_toolbox.admin import ReadOnlyTabularInline

from .dummy.models import Child
from .dummy.models import Parent

pytestmark = pytest.mark.django_db


class ChildStackedInline(ReadOnlyStackedInline[Child, Parent]):
    model = Child
    fk_name = "parent"
    fields = ["bar"]


class ChildTabularInline(ReadOnlyTabularInline[Child, Parent]):
    model = Child
    fk_name = "parent"
    fields = ["bar"]


@pytest.mark.parametrize("inline_class", [ChildStackedInline, ChildTabularInline])
class TestReadOnlyInline:
    def test_inline_configuration(self, inline_class):
        inline = inline_class(Parent, admin.site)

        assert inline.model == Child
        assert inline.fk_name == "parent"
        assert inline.fields == ["bar"]

    def test_readonly_fields(self, rf, admin_user, inline_class):
        request = rf.get("/")
        request.user = admin_user

        inline = inline_class(Parent, admin.site)
        readonly_fields = inline.get_readonly_fields(request)

        assert set(inline.fields).issubset(readonly_fields)

    def test_inline_permissions(self, rf, admin_user, inline_class):
        request = rf.get("/")
        request.user = admin_user

        inline = inline_class(Parent, admin.site)
        parent = baker.make("dummy.Parent")

        assert inline.has_add_permission(request, parent) is False
        assert inline.has_change_permission(request, parent) is False
        assert inline.can_delete is False
        assert inline.extra == 0

    def test_fields_subset_of_model_fields(self, inline_class):
        inline = inline_class(Parent, admin.site)
        model_fields = set(f.name for f in Child._meta.fields)

        assert set(inline.fields).issubset(model_fields)



from __future__ import annotations

import pytest
from django.core.exceptions import ImproperlyConfigured
from django.http import QueryDict
from django_tables2.views import SingleTableMixin
from model_bakery import baker
from neapolitan.views import Role

from .models import Bookmark
from .views import BookmarkTableOrderedView
from .views import BookmarkTableView
from .views import BookmarkView


def test_get_context_data_no_object():
    context = BookmarkView().get_context_data()

    assert "list_view_url" in context.keys()
    assert "delete_view_url" not in context.keys()
    assert "detail_view_url" not in context.keys()
    assert "update_view_url" not in context.keys()


def test_get_context_data_with_object(db):
    bookmark = baker.make(Bookmark)

    context = BookmarkView(object=bookmark).get_context_data()

    assert "list_view_url" in context.keys()
    assert "delete_view_url" in context.keys()
    assert "detail_view_url" in context.keys()
    assert "update_view_url" in context.keys()


@pytest.mark.parametrize(
    "role,expected",
    [
        (Role.DETAIL, ["url", "title"]),
        (Role.LIST, ["url"]),
        (Role.CREATE, ["url", "title", "note"]),
        (Role.UPDATE, ["url", "title", "note"]),
        (Role.DELETE, ["url", "title", "note"]),
    ],
)
def test_get_fields(role, expected):
    fields = BookmarkView(role=role).get_fields()

    assert fields == expected


@pytest.mark.parametrize(
    "role",
    [
        Role.DETAIL,
        Role.LIST,
        Role.CREATE,
        Role.UPDATE,
        Role.DELETE,
    ],
)
def test_get_fields_only(role):
    fields = BookmarkView(role=role, detail_fields=None, list_fields=None).get_fields()

    assert fields == BookmarkView.fields


@pytest.mark.parametrize(
    "role",
    [
        Role.DETAIL,
        Role.LIST,
        Role.CREATE,
        Role.UPDATE,
        Role.DELETE,
    ],
)
def test_get_fields_override(role):
    class BookmarkViewGetFieldsOverride(BookmarkView):
        def get_fields(self):
            return BookmarkView.fields

    fields = BookmarkViewGetFieldsOverride(
        role=role, fields=None, detail_fields=None, list_fields=None
    ).get_fields()

    assert fields == BookmarkView.fields


@pytest.mark.parametrize(
    "role",
    [
        Role.DETAIL,
        Role.LIST,
        Role.CREATE,
        Role.UPDATE,
        Role.DELETE,
    ],
)
def test_get_fields_no_fields(role):
    with pytest.raises(ImproperlyConfigured):
        BookmarkView(
            role=role, fields=None, detail_fields=None, list_fields=None
        ).get_fields()


def test_get_detail_fields():
    fields = BookmarkView().get_detail_fields()

    assert fields != BookmarkView.fields
    assert fields == BookmarkView.detail_fields


def test_get_detail_fields_override():
    class BookmarkViewGetDetailFieldsOverride(BookmarkView):
        def get_detail_fields(self):
            return BookmarkView.fields

    fields = BookmarkViewGetDetailFieldsOverride().get_detail_fields()

    assert fields == BookmarkView.fields
    assert fields != BookmarkView.detail_fields


def test_get_list_fields():
    fields = BookmarkView().get_list_fields()

    assert fields != BookmarkView.fields
    assert fields == BookmarkView.list_fields


def test_get_list_fields_override():
    class BookmarkViewGetListFieldsOverride(BookmarkView):
        def get_list_fields(self):
            return BookmarkView.fields

    fields = BookmarkViewGetListFieldsOverride().get_list_fields()

    assert fields == BookmarkView.fields
    assert fields != BookmarkView.list_fields


@pytest.mark.parametrize(
    "klass,expected",
    [
        (BookmarkView, False),
        (BookmarkTableView, True),
    ],
)
def test_as_view_table_class(klass, expected):
    view = klass.as_view(role=Role.LIST)

    assert issubclass(view.view_class, SingleTableMixin) is expected


@pytest.mark.parametrize(
    "klass,expected",
    [
        (BookmarkView, False),
        (BookmarkTableView, True),
    ],
)
def test_get_context_data_table(klass, expected, rf, db):
    View = klass.as_view(role=Role.LIST, object=baker.make(Bookmark)).view_class
    request = rf.get(Role.LIST.maybe_reverse(View))

    context = View(request=request).get_context_data()

    assert ("table" in context.keys()) is expected


@pytest.mark.parametrize(
    "htmx,expected",
    [
        (True, "neapolitan/object_list.html#object-list"),
        (False, "neapolitan/object_list.html"),
    ],
)
def test_get_template_names(htmx, expected, rf):
    request = rf.get(Role.LIST.maybe_reverse(BookmarkView))
    request.htmx = htmx

    view = BookmarkView(role=Role.LIST, **Role.LIST.extra_initkwargs())
    view.setup(request)

    template_names = view.get_template_names()

    assert expected in template_names


def test_get_template_names_no_htmx(rf):
    request = rf.get(Role.LIST.maybe_reverse(BookmarkView))

    view = BookmarkView(role=Role.LIST, **Role.LIST.extra_initkwargs())
    view.setup(request)

    template_names = view.get_template_names()

    assert "neapolitan/object_list.html" in template_names
    assert "neapolitan/object_list.html#object-list" not in template_names


def test_table_view_ordered(client, db):
    baker.make(Bookmark, _quantity=3)

    response = client.get(Role.LIST.maybe_reverse(BookmarkTableOrderedView))

    assert response.status_code == 200


def test_filterset_primary_fields(rf):
    class BookmarkFilterSetPrimaryFieldsView(BookmarkView):
        filterset_fields = ["url", "title", "note"]
        filterset_primary_fields = ["url"]

    request = rf.get(Role.LIST.maybe_reverse(BookmarkView))

    view = BookmarkFilterSetPrimaryFieldsView(
        role=Role.LIST, **Role.LIST.extra_initkwargs()
    )
    view.setup(request)

    filterset = view.get_filterset()

    assert set(filterset.primary_fields) == {"url"}
    assert set(filterset.secondary_fields) == {"title", "note"}


def test_filterset_extra_methods(rf):
    class BookmarkFilterSetExtraMethodsView(BookmarkView):
        filterset_fields = ["url", "title", "note"]

    request = rf.get(
        Role.LIST.maybe_reverse(BookmarkView), data=QueryDict("url=example.com")
    )

    view = BookmarkFilterSetExtraMethodsView(
        role=Role.LIST, **Role.LIST.extra_initkwargs()
    )
    view.setup(request)

    filterset = view.get_filterset()

    filterset.form.is_valid()

    assert hasattr(filterset, "is_active")
    assert hasattr(filterset.__class__, "active_filters")


def test_filterset_is_active(rf):
    class BookmarkFilterSetIsActiveView(BookmarkView):
        filterset_fields = ["url", "title", "note"]

    request = rf.get(
        Role.LIST.maybe_reverse(BookmarkView), data=QueryDict("url=example.com")
    )

    view = BookmarkFilterSetIsActiveView(role=Role.LIST, **Role.LIST.extra_initkwargs())
    view.setup(request)

    filterset = view.get_filterset()

    filterset.form.is_valid()

    assert callable(filterset.is_active)
    assert filterset.is_active() is True


def test_filterset_active_filters(rf):
    class BookmarkFilterSetActiveFiltersView(BookmarkView):
        filterset_fields = ["url", "title", "note"]

    request = rf.get(
        Role.LIST.maybe_reverse(BookmarkView), data=QueryDict("url=example.com")
    )

    view = BookmarkFilterSetActiveFiltersView(
        role=Role.LIST, **Role.LIST.extra_initkwargs()
    )
    view.setup(request)

    filterset = view.get_filterset()

    filterset.form.is_valid()

    assert isinstance(filterset.active_filters, dict)
    assert filterset.active_filters == {"url": "example.com"}


def test_filterset_extra_methods_no_filters(rf):
    class BookmarkFilterSetExtraMethodsView(BookmarkView):
        filterset_fields = ["url", "title", "note"]

    request = rf.get(Role.LIST.maybe_reverse(BookmarkView))

    view = BookmarkFilterSetExtraMethodsView(
        role=Role.LIST, **Role.LIST.extra_initkwargs()
    )
    view.setup(request)

    filterset = view.get_filterset()

    filterset.form.is_valid()

    assert filterset.is_active() is False
    assert filterset.active_filters == {}


@pytest.mark.parametrize(
    "role",
    [
        Role.DETAIL,
        Role.LIST,
        Role.CREATE,
        Role.UPDATE,
        Role.DELETE,
    ],
)
def test_get_role_context_data(role):
    class BookmarkGetRoleContextDataView(BookmarkView):
        def get_detail_context_data(self, context, **kwargs):
            return {"detail": "value"}

        def get_list_context_data(self, context, **kwargs):
            return {"list": "value"}

        def get_create_context_data(self, context, **kwargs):
            return {"create": "value"}

        def get_update_context_data(self, context, **kwargs):
            return {"update": "value"}

        def get_delete_context_data(self, context, **kwargs):
            return {"delete": "value"}

    view = BookmarkGetRoleContextDataView()
    view.role = role

    result = view.get_role_context_data({"context": "data"})

    assert result == {role.value: "value"}


@pytest.mark.parametrize(
    "role",
    [
        Role.DETAIL,
        Role.LIST,
        Role.CREATE,
        Role.UPDATE,
        Role.DELETE,
    ],
)
def test_get_role_context_data_no_arg(role):
    class BookmarkGetRoleContextDataNoArgView(BookmarkView):
        def get_detail_context_data(self, **kwargs):
            return {"detail": "value"}

        def get_list_context_data(self, **kwargs):
            return {"list": "value"}

        def get_create_context_data(self, **kwargs):
            return {"create": "value"}

        def get_update_context_data(self, **kwargs):
            return {"update": "value"}

        def get_delete_context_data(self, **kwargs):
            return {"delete": "value"}

    view = BookmarkGetRoleContextDataNoArgView()
    view.role = role

    result = view.get_role_context_data({"context": "data"})

    assert result == {role.value: "value"}


@pytest.mark.parametrize(
    "role",
    [
        Role.DETAIL,
        Role.LIST,
        Role.CREATE,
        Role.UPDATE,
        Role.DELETE,
    ],
)
def test_get_role_context_data_nonexistent(role):
    view = BookmarkView()
    view.role = role

    result = view.get_role_context_data({})

    assert result == {}


@pytest.mark.parametrize(
    "role",
    [
        Role.DETAIL,
        Role.LIST,
        Role.CREATE,
        Role.UPDATE,
        Role.DELETE,
    ],
)
def test_get_role_context_data_incorrect_return(role):
    class BookmarkRoleContextDataView(BookmarkView):
        def get_detail_context_data(self, context, **kwargs):
            return "not a dict"

        def get_list_context_data(self, context, **kwargs):
            return "not a dict"

        def get_create_context_data(self, context, **kwargs):
            return "not a dict"

        def get_update_context_data(self, context, **kwargs):
            return "not a dict"

        def get_delete_context_data(self, context, **kwargs):
            return "not a dict"

    view = BookmarkRoleContextDataView()
    view.role = role

    with pytest.raises(ValueError):
        view.get_role_context_data({})



from __future__ import annotations

from django_tables2 import tables

from django_twc_toolbox.crud.views import CRUDView

from .models import Bookmark


class BookmarkView(CRUDView):
    model = Bookmark
    fields = ["url", "title", "note"]

    detail_fields = ["url", "title"]
    list_fields = ["url"]


class BookmarkTable(tables.Table):
    class Meta:
        model = Bookmark


class BookmarkTableView(BookmarkView):
    table_class = BookmarkTable
    url_base = "bookmarktable"


class BookmarkTableOrdered(tables.Table):
    class Meta:
        model = Bookmark
        order_by = "title"


class BookmarkTableOrderedView(BookmarkView):
    table_class = BookmarkTableOrdered
    paginate_by = 1
    url_base = "bookmarktableordered"


urlpatterns = [
    *BookmarkView.get_urls(),
    *BookmarkTableView.get_urls(),
    *BookmarkTableOrderedView.get_urls(),
]



from __future__ import annotations

import pytest
from django.test import override_settings


@pytest.fixture(autouse=True)
def crud_settings():
    with override_settings(ROOT_URLCONF="tests.test_crud.views"):
        yield



from __future__ import annotations

import sys
from typing import Generic

from django.contrib import admin
from django.contrib.admin.options import BaseModelAdmin
from django.http import HttpRequest

from django_twc_toolbox._types import ChildModelT
from django_twc_toolbox._types import ListOrTuple
from django_twc_toolbox._types import ParentModelT

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override  # pyright: ignore[reportUnreachable]


class ReadOnlyModelAdmin(
    Generic[ChildModelT, ParentModelT], BaseModelAdmin[ChildModelT]
):
    can_delete = False
    extra = 0

    @override
    def get_readonly_fields(
        self, request: HttpRequest, obj: ChildModelT | None = None
    ) -> ListOrTuple[str]:
        return [f.name for f in self.model._meta.fields]

    @override
    def has_add_permission(  # type: ignore[override]
        self, request: HttpRequest, obj: ParentModelT | None
    ) -> bool:
        return False

    @override
    def has_change_permission(  # type: ignore[override]
        self, request: HttpRequest, obj: ParentModelT | None = None
    ) -> bool:
        return False


class ReadOnlyStackedInline(  # pyright: ignore[reportUnsafeMultipleInheritance]
    ReadOnlyModelAdmin[ChildModelT, ParentModelT],
    admin.StackedInline[ChildModelT, ParentModelT],
): ...


class ReadOnlyTabularInline(  # pyright: ignore[reportUnsafeMultipleInheritance]
    ReadOnlyModelAdmin[ChildModelT, ParentModelT],
    admin.TabularInline[ChildModelT, ParentModelT],
): ...



from __future__ import annotations

from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Mapping
from collections.abc import Sequence
from urllib.parse import quote

from django.http import HttpResponse
from django.urls import reverse as dj_reverse
from django.utils.functional import lazy
from django.utils.http import urlencode


def reverse(
    viewname: Callable[..., HttpResponse] | str | None,
    urlconf: str | None = None,
    args: Sequence[object] | None = None,
    kwargs: dict[str, object] | None = None,
    current_app: str | None = None,
    # https://github.com/typeddjango/django-stubs/blob/c7d734af5703712ad9bb78ff5be902c760ccb637/django-stubs/utils/http.pyi#L13-L17
    query: Mapping[str, str | bytes | int | Iterable[str | bytes | int]]
    | Iterable[tuple[str, str | bytes | int | Iterable[str | bytes | int]]]
    | None = None,
    fragment: str | None = None,
) -> str:
    url = dj_reverse(
        viewname, urlconf=urlconf, args=args, kwargs=kwargs, current_app=current_app
    )
    if query:
        url += f"?{urlencode(query)}"
    if fragment:
        url += f"#{quote(fragment)}"
    return url


reverse_lazy = lazy(reverse, str)



from __future__ import annotations

from collections.abc import Generator
from collections.abc import Sequence
from typing import cast

from django import template
from django.db import models
from neapolitan.views import Role

from django_twc_toolbox.crud.views import CRUDView

register = template.Library()


def action_links(view: CRUDView, object: models.Model):
    actions = {
        "detail": {
            "url": Role.DETAIL.maybe_reverse(view, object),
            "text": "View",
        },
        "update": {
            "url": Role.UPDATE.maybe_reverse(view, object),
            "text": "Edit",
        },
        "delete": {
            "url": Role.DELETE.maybe_reverse(view, object),
            "text": "Delete",
        },
    }
    return actions


@register.inclusion_tag("neapolitan/partial/detail.html")
def object_detail(object: models.Model, view: CRUDView):
    """
    Renders a detail view of an object with the given fields.

    Inclusion tag usage::

        {% object_detail object view %}

    Template: ``neapolitan/partial/detail.html`` - Will render a table of the
    object's fields.
    """

    fields = view.get_fields()

    def iter() -> Generator[tuple[str, str], None, None]:
        for f in fields:
            mf = object._meta.get_field(f)
            yield (cast(str, mf.verbose_name), str(getattr(object, f)))  # type: ignore[union-attr]

    return {"object": iter()}


@register.inclusion_tag("neapolitan/partial/list.html")
def object_list(objects: Sequence[models.Model], view: CRUDView):
    """
    Renders a list of objects with the given fields.

    Inclusion tag usage::

        {% object_list objects view %}

    Template: ``neapolitan/partial/list.html`` — Will render a table of objects
    with links to view, edit, and delete views.
    """

    fields = view.get_fields()
    headers = [cast(str, objects[0]._meta.get_field(f).verbose_name) for f in fields]  # type: ignore[union-attr]
    object_list = [
        {
            "object": object,
            "fields": [{"name": f, "value": str(getattr(object, f))} for f in fields],
            "actions": action_links(view, object),
        }
        for object in objects
    ]

    return {
        "headers": headers,
        "object_list": object_list,
    }



from __future__ import annotations

import pytest
from django.test import override_settings
from django.urls import NoReverseMatch
from django.urls import path

from django_twc_toolbox.urls import reverse


def view(request):
    return "Hello World"


def view_with_arg(request, arg):
    return arg


urlpatterns = [
    path("view/", view, name="view"),
    path("view_with_arg/<int:arg>/", view_with_arg, name="view_with_arg"),
]


@pytest.fixture(autouse=True)
def urlconf():
    with override_settings(ROOT_URLCONF=__name__):
        yield


@pytest.mark.parametrize(
    "viewname,args,kwargs,query,fragment,expected",
    [
        ("view", None, None, None, None, "/view/"),
        (view, None, None, None, None, "/view/"),
        ("view_with_arg", (123,), None, None, None, "/view_with_arg/123/"),
        ("view_with_arg", None, {"arg": 123}, None, None, "/view_with_arg/123/"),
        ("view", None, None, {"param": "hello"}, None, "/view/?param=hello"),
        ("view", None, None, None, "fragment", "/view/#fragment"),
        (
            "view",
            None,
            None,
            {"param": "hello"},
            "fragment",
            "/view/?param=hello#fragment",
        ),
        (
            "view",
            None,
            None,
            {"param": "hello++"},
            "fragment&more",
            "/view/?param=hello%2B%2B#fragment%26more",
        ),
        (
            "view",
            None,
            None,
            {"params": ["hello", "world"]},
            None,
            "/view/?params=%5B%27hello%27%2C+%27world%27%5D",
        ),
        ("view", None, None, {"param": ""}, None, "/view/?param="),
    ],
)
def test_reverse(viewname, args, kwargs, query, fragment, expected):
    assert (
        reverse(viewname, args=args, kwargs=kwargs, query=query, fragment=fragment)
        == expected
    )


@pytest.mark.parametrize(
    "viewname,args,kwargs,query,fragment,expected",
    [
        ("view", None, {"unexpected": "param"}, None, None, NoReverseMatch),
        ("view_with_arg", None, None, None, None, NoReverseMatch),
        ("view_with_arg", "incorrect_type", None, None, None, NoReverseMatch),
        ("view_with_arg", None, {"arg": "incorrect_type"}, None, None, NoReverseMatch),
        ("view", None, None, {"param": None}, None, TypeError),
        ("nonexistent_view", None, None, None, None, NoReverseMatch),
    ],
)
def test_reverse_invalid(viewname, args, kwargs, query, fragment, expected):
    with pytest.raises(expected):
        reverse(viewname, args=args, kwargs=kwargs, query=query, fragment=fragment)



from collections.abc import Iterable

from django.db import models
from django.utils.safestring import SafeString
from neapolitan.views import CRUDView

def action_links(view: CRUDView, object: models.Model) -> SafeString: ...
def object_detail(
    object: models.Model, fields: list[str]
) -> dict[str, tuple[str, str]]: ...
def object_list(
    objects: Iterable[models.Model], view: CRUDView
) -> dict[str, list[str | dict[str, models.Model | list[str] | SafeString]]]: ...



from __future__ import annotations

from dataclasses import dataclass

from django.conf import settings

from django_twc_toolbox._types import override

DJANGO_TWC_TOOLBOX_SETTINGS_NAME = "DJANGO_TWC_TOOLBOX"


@dataclass(frozen=True)
class AppSettings:
    CACHE_TIME_ROBOTS_TXT = 60 * 60 * 24  # one day
    CACHE_TIME_SECURITY_TXT = 60 * 60 * 24  # one day
    TEMPLATE_404 = "404.html"
    TEMPLATE_500 = "500.html"
    TEMPLATE_ROBOTS_TXT = "robots.txt"
    TEMPLATE_SECURITY_TXT = ".well-known/security.txt"
    SENTRY_DISGARDED_METHODS = ["GET", "HEAD"]
    SENTRY_DISGARDED_PATHS = ["/health/"]
    SENTRY_PROFILE_RATE = 0.5
    SENTRY_TRACES_RATE = 0.5

    @override
    def __getattribute__(self, __name: str) -> object:
        user_settings = getattr(settings, DJANGO_TWC_TOOLBOX_SETTINGS_NAME, {})
        return user_settings.get(__name, super().__getattribute__(__name))  # pyright: ignore[reportAny]


app_settings = AppSettings()



from __future__ import annotations

import pytest
from model_bakery import baker

pytestmark = pytest.mark.django_db


def test_with_history():
    dummy = baker.make("dummy.ModelWithHistory")

    assert dummy.history.count() == 1


def test_save_without_history_kwarg():
    dummy = baker.make("dummy.ModelWithHistory")

    assert dummy.history.count() == 1

    dummy.name = "New name"
    dummy.save(without_history=True)

    assert dummy.history.count() == 1

    dummy.name = "I'm not a dummy"
    dummy.save()

    assert dummy.history.count() == 2


def test_save_without_history_method():
    dummy = baker.make("dummy.ModelWithHistory")

    assert dummy.history.count() == 1

    dummy.name = "You're a dummy"
    dummy.save_without_history()

    assert dummy.history.count() == 1
